import copy
import logging
import signal
import threading
from typing import Dict, Optional, Set

from . import Incident, instantiate_variable

logger = logging.getLogger(__name__)


class Monitor:
    def __init__(self, interval: Optional[float] = None):
        self._interval = interval
        self._constants = set()
        self._variables = set()
        self._incidents = set()
        self._locals = {}
        self._running = None

        if self._interval:
            signal.signal(signal.SIGINT, self.exit)
            signal.signal(signal.SIGTERM, self.exit)
            logger.info(f"Monitoring is set for {interval} seconds.")

    def exit(self, *args):
        logger.info(f"Exiting")
        self._running.set()

    def add_constants(self, json: Dict) -> int:
        return self._parse_variables(json, self._constants)

    def add_variables(self, json: Dict) -> int:
        return self._parse_variables(json, self._variables)

    def add_incidents(self, json: Dict) -> int:
        count = 0
        for key, json in json.items():
            incident = Incident(key, json)
            if incident in self._incidents:
                raise Exception(f"Incident '{key}' already exists")
            self._incidents.add(incident)
            logger.info(f"Incident '{incident.name}' is configured")
            count += 1
        return count

    def start(self) -> None:
        for v in self._constants:
            self._locals[v.name] = v.value
            logger.info(f"Constant '{v.name}' holds {v.value}")

        if self._interval:
            thread: threading.Thread = None
            self._running = threading.Event()
            while not self._running.wait(self._interval):
                if thread:
                    thread.join()
                if not self._running.is_set():
                    thread = threading.Thread(target=self.fetch_and_evalute)
                    thread.start()
        else:
            self.fetch_and_evalute()

    def fetch_and_evalute(self) -> None:
        locals = copy.deepcopy(self._locals)
        for v in self._variables:
            value = v.new_value()
            last_value = v.last_value
            locals[v.name] = value
            logger.info(f"'{v.name}' is loaded and holds {value}")
            locals["last_" + v.name] = last_value
            logger.info(f"'last_{v.name}' is loaded and holds '{last_value}'")
        logger.debug("Reloaded all variables")

        for incident in self._incidents:
            logger.debug(f"Evaluating '{incident.name}' incident")
            incident.escalated(locals)
        logger.debug("Evaludated all incidents")

    def _parse_variables(self, json: Dict, containor: Set) -> int:
        count = 0
        for key, json in json.items():
            variable = instantiate_variable(key, json)
            if variable in self._constants or variable in self._variables:
                raise Exception(f"Variable '{key}' already exists")
            containor.add(variable)
            logger.info(f"Variable '{variable.name}' is configured")
            count += 1
        return count



import argparse
import logging
import json
import signal
import sys
import os

from . import Monitor

logger = logging.getLogger(__name__)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-c",
        "--config",
        dest="config",
        type=argparse.FileType("r"),
        required=True,
        nargs="+",
        help="configuration file to specify sysctl monitoring and their actions",
    )

    parser.add_argument(
        "-i",
        "--interval",
        dest="interval",
        type=float,
        required=False,
        help="iternal in second to re-evaludate rules",
    )

    parser.add_argument(
        "-l",
        "--log",
        dest="log",
        type=str,
        required=False,
        help="log file.  If not specified, logging is disabled",
    )

    parser.add_argument(
        "-d",
        "--debug",
        dest="debug",
        action="store_true",
        help="use debug level logging",
    )
    parser.set_defaults(debug=False)

    return parser.parse_args()


def analyze(args):
    if args.log:
        logging.basicConfig(
            filename=args.log, level=logging.DEBUG if args.debug else logging.INFO
        )
    else:
        logging.disable(logging.CRITICAL)

    m = Monitor(args.interval)
    for file in args.config:
        with file as json_file:
            setting = json.load(json_file)
            if "constants" in setting:
                logger.debug(f"Loading constants from '{file.name}'")
                m.add_constants(setting["constants"])
            if "variables" in setting:
                logger.debug(f"Loading variables from '{file.name}'")
                m.add_variables(setting["variables"])
            if "incidents" in setting:
                logger.debug(f"Loading incidents from '{file.name}'")
                m.add_incidents(setting["incidents"])
    m.start()


def main():
    args = parse_args()
    try:
        analyze(args)
    except KeyboardInterrupt:
        try:
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            sys.exit(0)
        except SystemExit:
            os._exit(0)

if __name__ == "__main__":
    main()



import logging
import os
import struct
import sys
from abc import ABC, abstractmethod
from typing import Any, Dict

import sysctl

logger = logging.getLogger(__name__)


class Variable(ABC):
    """
    A user defines a variable with its "name" and a way to fetch its value.
    """

    def __init__(self, name: str, params: Dict):
        assert name
        self._name = name
        self._value = None
        self._last_value = None

    def __hash__(self):
        return hash(self._name)

    def __eq__(self, other) -> bool:
        return self._name == other._name

    @property
    def name(self) -> str:
        return self._name

    @property
    def value(self) -> Any:
        return self._value

    @property
    def last_value(self) -> Any:
        return self._last_value

    def new_value(self) -> Any:
        self._last_value = self._value
        self._value = self._fetch_value()
        return self._value

    @abstractmethod
    def _fetch_value(self) -> Any:
        return self._value


class SyscmdVariable(Variable):
    """
    A user defines a variable with its "name" and sysctl name for its value.
    """

    def __init__(self, name: str, params: Dict):
        super().__init__(name, params)

        self._cmd = params["syscmd"]
        self._last_value = self._value = self._fetch_value()

    def _fetch_value(self) -> Any:
        return os.popen(self._cmd).read().strip()


TYPE = {
    # "node": None,
    "integer": "i",
    # "string": None,
    "int64_t": "q",
    # "opaque": None,
    "unsigned integer": "I",
    "long integer": "l",
    "long": "l",
    "unsigned long": "L",
    "uint64_t": "Q",
    "uint8_t": "B",
    "uint16_t": "H",
    "int8_t": "b",
    "int16_t": "h",
    "int32_t": "i",
    "uint32_t": "I",
}

VMTOTAL = [
    ("uint64_t", "t_vm"),  #     /* total virtual memory */
    ("uint64_t", "t_avm"),  #    /* active virtual memory */
    ("uint64_t", "t_rm"),  #     /* total real memory in use */
    ("uint64_t", "t_arm"),  #    /* active real memory */
    ("uint64_t", "t_vmshr"),  #  /* shared virtual memory */
    ("uint64_t", "t_avmshr"),  # /* active shared virtual memory */
    ("uint64_t", "t_rmshr"),  #  /* shared real memory */
    ("uint64_t", "t_armshr"),  # /* active shared real memory */
    ("uint64_t", "t_free"),  #   /* free memory pages */
    ("int16_t", "t_rq"),  #      /* length of the run queue */
    ("int16_t", "t_dw"),  #      /* threads in ``disk wait'' (neg priority) */
    ("int16_t", "t_pw"),  #      /* threads in page wait */
    ("int16_t", "t_sl"),  #      /* threads sleeping in core */
    ("int16_t", "t_sw"),  #      /* swapped out runnable/short block threads */
]

LOADAVG = [
    ("uint32_t", "1min"),
    ("uint32_t", "3min"),
    ("uint32_t", "15min"),
    ("long", "scale"),
]


class SysctlVariable(Variable):
    """
    A user defines a variable with its "name" and sysctl name for its value.
    """

    def __init__(self, name: str, params: Dict):
        super().__init__(name, params)

        self._sysctl_name = params["sysctl"]
        ctl = SysctlVariable._verify_sysctl(self._sysctl_name)
        self._last_value = self._value = SysctlVariable.extract(ctl)

    def _fetch_value(self) -> Any:
        # Sysctl caches and thus need a new instance each time
        ctl = sysctl.filter(self._sysctl_name)[0]
        return SysctlVariable.extract(ctl)

    @staticmethod
    def extract(ctl: "Sysctl") -> Any:
        if ctl.type == 10:  # CTLTYPE_U8
            offset = 1
        elif ctl.type == 11:  # CTLTYPE_U16
            offset = 2
        elif ctl.type == 15:  # CTLTYPE_U32
            offset = 4
        elif ctl.type == 5:  # CTLTYPE_U32
            return SysctlVariable.extract_struct(ctl)
        else:
            return ctl.value
        return int.from_bytes(ctl.value[0:offset], sys.byteorder, signed=False)

    @staticmethod
    def extract_loadavg(ctl: "Sysctl") -> Any:
        tmp = SysctlVariable.transform_struct(ctl, LOADAVG)
        scale = float(tmp["scale"])
        return (tmp["1min"] / scale, tmp["3min"] / scale, tmp["15min"] / scale)

    @staticmethod
    def transform_struct(ctl: "Sysctl", mapping: Dict) -> Dict:
        value = {}
        start = 0
        for field in mapping:
            format = TYPE[field[0]]
            offset = struct.calcsize(format)  # TODO - optimize
            bytes = ctl.value[start : start + offset]
            (value[field[1]],) = struct.unpack_from(format, ctl.value, start)
            start += offset
        return value

    @staticmethod
    def extract_struct(ctl: "Sysctl") -> Any:
        if ctl.name == "vm.vmtotal":
            mapping = VMTOTAL
        elif ctl.name == "vm.loadavg":
            return SysctlVariable.extract_loadavg(ctl)
        else:
            return ctl.value
        return SysctlVariable.transform_struct(ctl, mapping)

    @staticmethod
    def _verify_sysctl(name: str) -> sysctl.Sysctl:
        ctl = sysctl.filter(name)
        assert type(ctl) == list  # TODO: need to see if I get something else
        if len(ctl) != 1:
            raise Exception(f"{name} is not supported type")
        return ctl[0]


def instantiate_variable(name: str, params: Dict) -> Any:
    assert name

    if "sysctl" in params:
        return SysctlVariable(name, params)
    elif "syscmd" in params:
        return SyscmdVariable(name, params)
    else:
        raise Exception("Unknown variable type")
